"""
Module to handle different Django/Python version libraries.
"""
try:
    from collections import OrderedDict
except ImportError:  # Py2.6, fall back to Django's implementation
    from django.utils.datastructures import SortedDict as OrderedDict  # pragma: no cover


try:
    from django.utils import six
except ImportError:
    import six  # pragma: no cover


try:
    from django.utils.deconstruct import deconstructible
except ImportError:
    # just return a noop decorator
    def deconstructible(cls):
        return cls



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from djchoices import DjangoChoices, C, ChoiceItem


class PrivateAPITests(unittest.TestCase):

    def test_labels_dict(self):
        """
        Ref #33: IPython reads the Labels.keys() for pretty printing.
        """
        class Choices(DjangoChoices):
            a = ChoiceItem('a', 'A')
            b = ChoiceItem('b', 'B')

        self.assertEqual(Choices.labels.a, 'A')
        self.assertEqual(Choices.labels.b, 'B')
        with self.assertRaises(AttributeError):
            Choices.labels.c

        keys = Choices.labels.keys()
        self.assertEqual(set(keys), set(['a', 'b']))




import re

from django.core.exceptions import ValidationError

from .compat import deconstructible, OrderedDict, six


__all__ = ["ChoiceItem", "DjangoChoices", "C"]


# Support Functionality (Not part of public API)

class Labels(dict):

    def __getattr__(self, name):
        result = dict.get(self, name, None)
        if result is not None:
            return result
        else:
            raise AttributeError("Label for field %s was not found." % name)

    def __setattr__(self, name, value):
        self[name] = value


class StaticProp(object):

    def __init__(self, value):
        self.value = value

    def __get__(self, obj, objtype):
        return self.value

# End Support Functionality


class ChoiceItem(object):
    """
    Describes a choice item.

    The label is usually the field name so label can normally be left blank.
    Set a label if you need characters that are illegal in a python identifier
    name (ie: "DVD/Movie").
    """
    order = 0

    def __init__(self, value=None, label=None, order=None):
        self.value = value
        self.label = label

        if order:
            self.order = order
        else:
            ChoiceItem.order += 1
            self.order = ChoiceItem.order

# Shorter convenience alias.
C = ChoiceItem


class DjangoChoicesMeta(type):
    """
    Metaclass that writes the choices class.
    """
    name_clean = re.compile(r"_+")

    def __new__(cls, name, bases, attrs):
        fields = {}
        labels = Labels()
        values = {}
        choices = []

        # Get all the fields from parent classes.
        parents = [b for b in bases if isinstance(b, DjangoChoicesMeta)]
        for kls in parents:
            for field_name in kls._fields:
                fields[field_name] = kls._fields[field_name]

        # Get all the fields from this class.
        for field_name in attrs:
            val = attrs[field_name]
            if isinstance(val, ChoiceItem):
                fields[field_name] = val

        fields = OrderedDict(sorted(fields.items(), key=lambda x: x[1].order))

        for field_name in fields:
            val = fields[field_name]
            if isinstance(val, ChoiceItem):
                if val.label is not None:
                    label = val.label
                else:
                    # TODO: mark translatable by default?
                    label = cls.name_clean.sub(" ", field_name)

                val0 = label if val.value is None else val.value
                choices.append((val0, label))
                attrs[field_name] = StaticProp(val0)
                setattr(labels, field_name, label)
                values[val0] = label
            else:
                choices.append((field_name, val.choices))

        attrs["choices"] = StaticProp(tuple(choices))
        attrs["labels"] = labels
        attrs["values"] = values
        attrs["_fields"] = fields
        attrs["validator"] = ChoicesValidator(values)

        return super(DjangoChoicesMeta, cls).__new__(cls, name, bases, attrs)


@deconstructible
class ChoicesValidator(object):

    def __init__(self, values):
        self.values = values

    def __call__(self, value):
        if value not in self.values:
            raise ValidationError('Select a valid choice. %s is not '
                                  'one of the available choices.' % value)

    def __eq__(self, other):
        return isinstance(other, ChoicesValidator) and self.values == other.values

    def __ne__(self, other):
        return not (self == other)


class DjangoChoices(six.with_metaclass(DjangoChoicesMeta)):
    order = 0
    choices = ()
    labels = Labels()
    values = {}
    validator = None



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from djchoices import DjangoChoices, C, ChoiceItem

from .utils import has_new_migrations


class NumericTestClass(DjangoChoices):
    Item_0 = C(0)
    Item_1 = C(1)
    Item_2 = C(2)
    Item_3 = C(3)


class StringTestClass(DjangoChoices):
    empty = ChoiceItem("", "")
    One = ChoiceItem("O")
    Two = ChoiceItem("T")
    Three = ChoiceItem("H")


class SubClass1(NumericTestClass):
    Item_4 = C(4)
    Item_5 = C(5)


class SubClass2(SubClass1):
    Item_6 = C(6)
    Item_7 = C(7)


class EmptyValueClass(DjangoChoices):
    Option1 = ChoiceItem()
    Option2 = ChoiceItem()
    Option3 = ChoiceItem()


class DjangoChoices(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_numeric_class_values(self):
        self.assertEqual(NumericTestClass.Item_0, 0)
        self.assertEqual(NumericTestClass.Item_1, 1)
        self.assertEqual(NumericTestClass.Item_2, 2)
        self.assertEqual(NumericTestClass.Item_3, 3)

    def test_class_labels(self):
        self.assertEqual(StringTestClass.labels.empty, "")
        self.assertEqual(NumericTestClass.labels.Item_1, "Item 1")
        self.assertEqual(NumericTestClass.labels.Item_2, "Item 2")
        self.assertEqual(NumericTestClass.labels.Item_3, "Item 3")

    def test_class_labels_inherited(self):
        self.assertEqual(SubClass2.labels.Item_2, "Item 2")
        self.assertEqual(SubClass2.labels.Item_6, "Item 6")

    def test_class_values(self):
        self.assertEqual(SubClass1.values[SubClass1.Item_1], "Item 1")
        self.assertEqual(SubClass1.values[SubClass1.Item_4], "Item 4")
        self.assertEqual(SubClass1.values[SubClass1.Item_5], "Item 5")

    def test_numeric_class_order(self):
        choices = NumericTestClass.choices
        self.assertEqual(choices[0][0], 0)
        self.assertEqual(choices[1][0], 1)
        self.assertEqual(choices[2][0], 2)
        self.assertEqual(choices[3][0], 3)

    def test_string_class_values(self):
        self.assertEqual(StringTestClass.One, "O")
        self.assertEqual(StringTestClass.Two, "T")
        self.assertEqual(StringTestClass.Three, "H")

    def test_string_class_order(self):
        choices = StringTestClass.choices
        self.assertEqual(choices[0][0], "")
        self.assertEqual(choices[1][0], "O")
        self.assertEqual(choices[2][0], "T")
        self.assertEqual(choices[3][0], "H")

    def test_sub_class_level_1_choices(self):
        choices = SubClass1.choices
        self.assertEqual(choices[0][0], 0)
        self.assertEqual(choices[4][0], 4)
        self.assertEqual(choices[5][0], 5)

    def test_sub_class_level_1_values(self):
        self.assertEqual(SubClass1.Item_1, 1)
        self.assertEqual(SubClass1.Item_4, 4)
        self.assertEqual(SubClass1.Item_5, 5)

    def test_sub_class_level_2_choices(self):
        choices = SubClass2.choices
        self.assertEqual(choices[0][0], 0)
        self.assertEqual(choices[4][0], 4)
        self.assertEqual(choices[6][0], 6)
        self.assertEqual(choices[7][0], 7)

    def test_sub_class_level_2_values(self):
        self.assertEqual(SubClass2.Item_1, 1)
        self.assertEqual(SubClass2.Item_5, 5)
        self.assertEqual(SubClass2.Item_6, 6)
        self.assertEqual(SubClass2.Item_7, 7)

    def test_sub_class_name(self):
        self.assertEqual(NumericTestClass.__name__, "NumericTestClass")
        self.assertEqual(SubClass2.__name__, "SubClass2")

    def test_numeric_class_validator(self):
        from django.core.exceptions import ValidationError

        self.assertEqual(None, NumericTestClass.validator(1))
        self.assertEqual(None, NumericTestClass.validator(2))
        self.assertEqual(None, NumericTestClass.validator(3))

        self.assertRaises(ValidationError, NumericTestClass.validator, 4)
        self.assertRaises(ValidationError, NumericTestClass.validator, 5)
        self.assertRaises(ValidationError, NumericTestClass.validator, 6)
        self.assertRaises(ValidationError, NumericTestClass.validator, 7)

    def test_validation_error_message(self):
        from django.core.exceptions import ValidationError

        message = ("Select a valid choice. 4 is not "
                   "one of the available choices.")

        self.assertRaisesRegexp(ValidationError, message, 
                                NumericTestClass.validator, 4)

    def test_subclass1_validator(self):
        from django.core.exceptions import ValidationError

        self.assertEqual(None, SubClass1.validator(1))
        self.assertEqual(None, SubClass1.validator(2))
        self.assertEqual(None, SubClass1.validator(3))
        self.assertEqual(None, SubClass1.validator(4))
        self.assertEqual(None, SubClass1.validator(5))

        self.assertRaises(ValidationError, SubClass1.validator, 6)
        self.assertRaises(ValidationError, SubClass1.validator, 7)

    def test_subclass_2_validator(self):
        from django.core.exceptions import ValidationError

        self.assertEqual(None, SubClass2.validator(1))
        self.assertEqual(None, SubClass2.validator(2))
        self.assertEqual(None, SubClass2.validator(3))
        self.assertEqual(None, SubClass2.validator(4))
        self.assertEqual(None, SubClass2.validator(5))
        self.assertEqual(None, SubClass2.validator(6))
        self.assertEqual(None, SubClass2.validator(7))

        self.assertRaises(ValidationError, SubClass2.validator, 8)

    def test_empty_value_class(self):
        choices = EmptyValueClass.choices
        self.assertEqual(choices[0][0], "Option1")
        self.assertEqual(choices[1][0], "Option2")
        self.assertEqual(choices[2][0], "Option3")

    @unittest.skipUnless(*has_new_migrations())
    def test_deconstructible_validator(self):
        deconstructed = NumericTestClass.validator.deconstruct()
        self.assertEqual(deconstructed, (
            'djchoices.choices.ChoicesValidator', (NumericTestClass.values,), {}
        ))



from pkg_resources import get_distribution

from djchoices.choices import ChoiceItem, DjangoChoices, C

__version__ = get_distribution('django-choices').version

__all__ = ["ChoiceItem", "DjangoChoices", "C"]
